#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
from unittest import mock
from unittest.mock import Mock

import numpy as np
import pandas as pd
from ax.core.arm import Arm
from ax.core.data import Data
from ax.core.metric import Metric
from ax.core.objective import Objective, ScalarizedObjective
from ax.core.observation import (
    ObservationData,
    ObservationFeatures,
    observations_from_data,
)
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import FixedParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UserInputError
from ax.modelbridge.base import (
    clamp_observation_features,
    gen_arms,
    GenResults,
    ModelBridge,
    unwrap_observation_data,
)
from ax.modelbridge.registry import Models
from ax.modelbridge.transforms.log import Log
from ax.models.base import Model
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment_with_multi_objective,
    get_experiment,
    get_experiment_with_repeated_arms,
    get_non_monolithic_branin_moo_data,
    get_optimization_config,
    get_optimization_config_no_constraints,
    get_search_space_for_range_value,
    get_search_space_for_range_values,
    get_search_space_for_value,
)
from ax.utils.testing.modeling_stubs import (
    get_experiment_for_value,
    get_observation1,
    get_observation1trans,
    get_observation2,
    get_observation2trans,
    get_observation_status_quo0,
    get_observation_status_quo1,
    transform_1,
    transform_2,
)


class BaseModelBridgeTest(TestCase):
    @mock.patch(
        "ax.modelbridge.base.observations_from_data",
        autospec=True,
        return_value=([get_observation1(), get_observation2()]),
    )
    @mock.patch(
        "ax.modelbridge.base.gen_arms",
        autospec=True,
        return_value=([Arm(parameters={})], None),
    )
    @mock.patch("ax.modelbridge.base.ModelBridge._fit", autospec=True)
    def test_ModelBridge(
        self, mock_fit: Mock, mock_gen_arms: Mock, mock_observations_from_data: Mock
    ) -> None:
        # Test that on init transforms are stored and applied in the correct order
        transforms = [transform_1, transform_2]
        exp = get_experiment_for_value()
        ss = get_search_space_for_value()
        modelbridge = ModelBridge(
            search_space=ss,
            model=Model(),
            # pyre-fixme[6]: For 3rd param expected
            #  `Optional[List[Type[Transform]]]` but got `List[Type[Union[transform_1,
            #  transform_2]]]`.
            transforms=transforms,
            experiment=exp,
            # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
            data=0,
        )
        self.assertFalse(
            modelbridge._experiment_has_immutable_search_space_and_opt_config
        )
        self.assertEqual(
            list(modelbridge.transforms.keys()), ["Cast", "transform_1", "transform_2"]
        )
        fit_args = mock_fit.mock_calls[0][2]
        self.assertTrue(fit_args["search_space"] == get_search_space_for_value(8.0))
        self.assertTrue(fit_args["observations"] == [])
        self.assertTrue(mock_observations_from_data.called)

        # Test prediction on out of design features.
        modelbridge._predict = mock.MagicMock(
            "ax.modelbridge.base.ModelBridge._predict",
            autospec=True,
            side_effect=ValueError("Out of Design"),
        )
        # This point is in design, and thus failures in predict are legitimate.
        with mock.patch.object(
            ModelBridge, "model_space", return_value=get_search_space_for_range_values
        ):
            with self.assertRaises(ValueError):
                modelbridge.predict([get_observation2().features])

        # This point is out of design, and not in training data.
        with self.assertRaises(ValueError):
            modelbridge.predict([get_observation_status_quo0().features])

        # Now it's in the training data.
        with mock.patch.object(
            ModelBridge,
            "get_training_data",
            return_value=[get_observation_status_quo0()],
        ):
            # Return raw training value.
            self.assertEqual(
                modelbridge.predict([get_observation_status_quo0().features]),
                unwrap_observation_data([get_observation_status_quo0().data]),
            )

        # Test that transforms are applied correctly on predict
        modelbridge._predict = mock.MagicMock(
            "ax.modelbridge.base.ModelBridge._predict",
            autospec=True,
            return_value=[get_observation2trans().data],
        )
        modelbridge.predict([get_observation2().features])
        # Observation features sent to _predict are un-transformed afterwards
        # pyre-fixme[16]: Callable `_predict` has no attribute `assert_called_with`.
        modelbridge._predict.assert_called_with([get_observation2().features])

        # Check that _single_predict is equivalent here.
        modelbridge._single_predict([get_observation2().features])
        # Observation features sent to _predict are un-transformed afterwards
        modelbridge._predict.assert_called_with([get_observation2().features])

        # Test transforms applied on gen
        modelbridge._gen = mock.MagicMock(
            "ax.modelbridge.base.ModelBridge._gen",
            autospec=True,
            return_value=GenResults(
                observation_features=[get_observation1trans().features], weights=[2]
            ),
        )
        oc = OptimizationConfig(objective=Objective(metric=Metric(name="test_metric")))
        modelbridge._set_kwargs_to_save(
            model_key="TestModel", model_kwargs={}, bridge_kwargs={}
        )
        # Test input error when generating 0 candidates.
        with self.assertRaisesRegex(UserInputError, "Attempted to generate"):
            modelbridge.gen(n=0)
        gr = modelbridge.gen(
            n=1,
            search_space=get_search_space_for_value(),
            optimization_config=oc,
            pending_observations={"a": [get_observation2().features]},
            fixed_features=ObservationFeatures({"x": 5}),
        )
        self.assertEqual(gr._model_key, "TestModel")
        # pyre-fixme[16]: Callable `_gen` has no attribute `assert_called_with`.
        modelbridge._gen.assert_called_with(
            n=1,
            search_space=SearchSpace([FixedParameter("x", ParameterType.FLOAT, 8.0)]),
            optimization_config=oc,
            pending_observations={"a": [get_observation2trans().features]},
            fixed_features=ObservationFeatures({"x": 36}),
            model_gen_options=None,
        )
        mock_gen_arms.assert_called_with(
            arms_by_signature={}, observation_features=[get_observation1().features]
        )

        # Gen with no pending observations and no fixed features
        modelbridge.gen(
            n=1, search_space=get_search_space_for_value(), optimization_config=None
        )
        modelbridge._gen.assert_called_with(
            n=1,
            search_space=SearchSpace([FixedParameter("x", ParameterType.FLOAT, 8.0)]),
            optimization_config=None,
            pending_observations={},
            fixed_features=ObservationFeatures({}),
            model_gen_options=None,
        )

        # Gen with multi-objective optimization config.
        oc2 = OptimizationConfig(
            objective=ScalarizedObjective(
                metrics=[Metric(name="test_metric"), Metric(name="test_metric_2")]
            )
        )
        modelbridge.gen(
            n=1, search_space=get_search_space_for_value(), optimization_config=oc2
        )
        modelbridge._gen.assert_called_with(
            n=1,
            search_space=SearchSpace([FixedParameter("x", ParameterType.FLOAT, 8.0)]),
            optimization_config=oc2,
            pending_observations={},
            fixed_features=ObservationFeatures({}),
            model_gen_options=None,
        )

        # Test transforms applied on cross_validate
        modelbridge._cross_validate = mock.MagicMock(
            "ax.modelbridge.base.ModelBridge._cross_validate",
            autospec=True,
            return_value=[get_observation1trans().data],
        )
        cv_training_data = [get_observation2()]
        cv_test_points = [get_observation1().features]
        cv_predictions = modelbridge.cross_validate(
            cv_training_data=cv_training_data, cv_test_points=cv_test_points
        )
        # pyre-fixme[16]: Callable `_cross_validate` has no attribute
        #  `assert_called_with`.
        modelbridge._cross_validate.assert_called_with(
            search_space=SearchSpace([FixedParameter("x", ParameterType.FLOAT, 8.0)]),
            cv_training_data=[get_observation2trans()],
            cv_test_points=[get_observation1().features],  # untransformed after
        )
        self.assertTrue(cv_predictions == [get_observation1().data])

        # Test stored training data
        obs = modelbridge.get_training_data()
        self.assertTrue(obs == [get_observation1(), get_observation2()])
        self.assertEqual(modelbridge.metric_names, {"a", "b"})
        self.assertIsNone(modelbridge.status_quo)
        self.assertTrue(modelbridge.model_space == get_search_space_for_value())
        self.assertEqual(modelbridge.training_in_design, [False, False])

        with self.assertRaises(ValueError):
            modelbridge.training_in_design = [True, True, False]

        with self.assertRaises(ValueError):
            modelbridge.training_in_design = [True, True, False]

        # Test feature_importances
        with self.assertRaises(NotImplementedError):
            modelbridge.feature_importances("a")

        # Test transform observation features
        with mock.patch(
            "ax.modelbridge.base.ModelBridge._transform_observation_features",
            autospec=True,
        ) as mock_tr:
            modelbridge.transform_observation_features([get_observation2().features])
        mock_tr.assert_called_with(modelbridge, [get_observation2trans().features])

    @mock.patch(
        "ax.modelbridge.base.observations_from_data",
        autospec=True,
        return_value=([get_observation1(), get_observation2()]),
    )
    def test_ood_gen(self, _) -> None:
        # Test fit_out_of_design by returning OOD candidats
        exp = get_experiment_for_value()
        ss = SearchSpace([RangeParameter("x", ParameterType.FLOAT, 0.0, 1.0)])
        modelbridge = ModelBridge(
            search_space=ss,
            model=Model(),
            transforms=[],
            experiment=exp,
            # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
            data=0,
            fit_out_of_design=True,
        )
        obs = ObservationFeatures(parameters={"x": 3.0})
        modelbridge._gen = mock.MagicMock(
            "ax.modelbridge.base.ModelBridge._gen",
            autospec=True,
            return_value=GenResults(observation_features=[obs], weights=[2]),
        )
        gr = modelbridge.gen(n=1)
        self.assertEqual(gr.arms[0].parameters, obs.parameters)

        # Test clamping arms by setting fit_out_of_design=False
        modelbridge = ModelBridge(
            search_space=ss,
            model=Model(),
            transforms=[],
            experiment=exp,
            # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
            data=0,
            fit_out_of_design=False,
        )
        obs = ObservationFeatures(parameters={"x": 3.0})
        modelbridge._gen = mock.MagicMock(
            "ax.modelbridge.base.ModelBridge._gen",
            autospec=True,
            return_value=GenResults(observation_features=[obs], weights=[2]),
        )
        gr = modelbridge.gen(n=1)
        self.assertEqual(gr.arms[0].parameters, {"x": 1.0})

    @mock.patch(
        "ax.modelbridge.base.observations_from_data",
        autospec=True,
        return_value=([get_observation1()]),
    )
    @mock.patch("ax.modelbridge.base.ModelBridge._fit", autospec=True)
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def testSetStatusQuo(self, mock_fit, mock_observations_from_data):
        # NOTE: If empty data object is not passed, observations are not
        # extracted, even with mock.
        modelbridge = ModelBridge(
            search_space=get_search_space_for_value(),
            model=0,
            experiment=get_experiment_for_value(),
            data=Data(),
            status_quo_name="1_1",
        )
        self.assertEqual(modelbridge.status_quo, get_observation1())

        # Alternatively, we can specify by features
        modelbridge = ModelBridge(
            get_search_space_for_value(),
            0,
            [],
            get_experiment_for_value(),
            # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
            0,
            status_quo_features=get_observation1().features,
        )
        self.assertEqual(modelbridge.status_quo, get_observation1())

        # Alternatively, we can specify on experiment
        # Put a dummy arm with SQ name 1_1 on the dummy experiment.
        exp = get_experiment_for_value()
        sq = Arm(name="1_1", parameters={"x": 3.0})
        exp._status_quo = sq
        # Check that we set SQ to arm 1_1
        # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
        modelbridge = ModelBridge(get_search_space_for_value(), 0, [], exp, 0)
        self.assertEqual(modelbridge.status_quo, get_observation1())

        # Errors if features and name both specified
        with self.assertRaises(ValueError):
            modelbridge = ModelBridge(
                get_search_space_for_value(),
                0,
                [],
                exp,
                # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
                0,
                status_quo_features=get_observation1().features,
                status_quo_name="1_1",
            )

        # Left as None if features or name don't exist
        modelbridge = ModelBridge(
            get_search_space_for_value(),
            0,
            [],
            exp,
            # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
            0,
            status_quo_name="1_0",
        )
        self.assertIsNone(modelbridge.status_quo)
        modelbridge = ModelBridge(
            get_search_space_for_value(),
            0,
            [],
            get_experiment_for_value(),
            # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
            0,
            status_quo_features=ObservationFeatures(parameters={"x": 3.0, "y": 10.0}),
        )
        self.assertIsNone(modelbridge.status_quo)

    @mock.patch(
        "ax.modelbridge.base.ModelBridge._gen",
        autospec=True,
    )
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_status_quo_for_non_monolithic_data(self, mock_gen):
        mock_gen.return_value = GenResults(
            observation_features=[
                ObservationFeatures(
                    parameters={"x1": float(i), "x2": float(i)}, trial_index=np.int64(1)
                )
                for i in range(5)
            ],
            # pyre-fixme[6]: For 2nd param expected `List[float]` but got `List[int]`.
            weights=[1] * 5,
        )
        exp = get_branin_experiment_with_multi_objective(with_status_quo=True)
        sobol = Models.SOBOL(search_space=exp.search_space)
        exp.new_batch_trial(sobol.gen(5)).set_status_quo_and_optimize_power(
            status_quo=exp.status_quo
        ).run()

        # create data where metrics vary in start and end times
        data = get_non_monolithic_branin_moo_data()
        with warnings.catch_warnings(record=True) as ws:
            bridge = ModelBridge(
                experiment=exp,
                data=data,
                model=Model(),
                search_space=exp.search_space,
            )
        # just testing it doesn't error
        bridge.gen(5)
        self.assertTrue(any("start_time" in str(w.message) for w in ws))
        self.assertTrue(any("end_time" in str(w.message) for w in ws))
        # pyre-fixme[16]: Optional type has no attribute `arm_name`.
        self.assertEqual(bridge.status_quo.arm_name, "status_quo")

    @mock.patch(
        "ax.modelbridge.base.observations_from_data",
        autospec=True,
        return_value=(
            [
                get_observation_status_quo0(),
                get_observation_status_quo1(),
                get_observation1(),
                get_observation2(),
            ]
        ),
    )
    @mock.patch("ax.modelbridge.base.ModelBridge._fit", autospec=True)
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def testSetStatusQuoMultipleObs(self, mock_fit, mock_observations_from_data):
        exp = get_experiment_with_repeated_arms(2)

        trial_index = 1
        status_quo_features = ObservationFeatures(
            # pyre-fixme[16]: `BaseTrial` has no attribute `status_quo`.
            parameters=exp.trials[trial_index].status_quo.parameters,
            # pyre-fixme[6]: For 2nd param expected `Optional[int64]` but got `int`.
            trial_index=trial_index,
        )
        modelbridge = ModelBridge(
            get_search_space_for_value(),
            0,
            [],
            exp,
            # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
            0,
            status_quo_features=status_quo_features,
        )
        # Check that for experiments with many trials the status quo is set
        # to the value of the status quo of the last trial.
        if len(exp.trials) >= 1:
            self.assertEqual(modelbridge.status_quo, get_observation_status_quo1())

    def test_transform_observations(self) -> None:
        """
        This functionality is unused, even in the subclass where it is implemented.
        """
        ss = get_search_space_for_value()
        modelbridge = ModelBridge(search_space=ss, model=Model())
        with self.assertRaises(NotImplementedError):
            modelbridge.transform_observations([])
        with self.assertRaises(NotImplementedError):
            modelbridge.transform_observations([])

    @mock.patch(
        "ax.modelbridge.base.observations_from_data",
        autospec=True,
        return_value=([get_observation1(), get_observation1()]),
    )
    @mock.patch("ax.modelbridge.base.ModelBridge._fit", autospec=True)
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def testSetTrainingDataDupFeatures(self, mock_fit, mock_observations_from_data):
        # Throws an error if repeated features in observations.
        with self.assertRaises(ValueError):
            ModelBridge(
                get_search_space_for_value(),
                0,
                [],
                get_experiment_for_value(),
                # pyre-fixme[6]: For 5th param expected `Optional[Data]` but got `int`.
                0,
                status_quo_name="1_1",
            )

    def testUnwrapObservationData(self) -> None:
        observation_data = [get_observation1().data, get_observation2().data]
        f, cov = unwrap_observation_data(observation_data)
        self.assertEqual(f["a"], [2.0, 2.0])
        self.assertEqual(f["b"], [4.0, 1.0])
        self.assertEqual(cov["a"]["a"], [1.0, 2.0])
        self.assertEqual(cov["b"]["b"], [4.0, 5.0])
        self.assertEqual(cov["a"]["b"], [2.0, 3.0])
        self.assertEqual(cov["b"]["a"], [3.0, 4.0])
        # Check that errors if metric mismatch
        od3 = ObservationData(
            metric_names=["a"], means=np.array([2.0]), covariance=np.array([[4.0]])
        )
        with self.assertRaises(ValueError):
            unwrap_observation_data(observation_data + [od3])

    def testGenArms(self) -> None:
        p1 = {"x": 0, "y": 1}
        p2 = {"x": 4, "y": 8}
        observation_features = [
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, int]`.
            ObservationFeatures(parameters=p1),
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool,
            #  float, int, str]]` but got `Dict[str, int]`.
            ObservationFeatures(parameters=p2),
        ]
        arms, candidate_metadata = gen_arms(observation_features=observation_features)
        self.assertEqual(arms[0].parameters, p1)
        self.assertIsNone(candidate_metadata)

        # pyre-fixme[6]: For 2nd param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, int]`.
        arm = Arm(name="1_1", parameters=p1)
        arms_by_signature = {arm.signature: arm}
        observation_features[0].metadata = {"some_key": "some_val_0"}
        observation_features[1].metadata = {"some_key": "some_val_1"}
        arms, candidate_metadata = gen_arms(
            observation_features=observation_features,
            arms_by_signature=arms_by_signature,
        )
        self.assertEqual(arms[0].name, "1_1")
        self.assertEqual(
            candidate_metadata,
            {
                arms[0].signature: {"some_key": "some_val_0"},
                arms[1].signature: {"some_key": "some_val_1"},
            },
        )

    @mock.patch(
        "ax.modelbridge.base.ModelBridge._gen",
        autospec=True,
        return_value=GenResults(
            observation_features=[get_observation1trans().features], weights=[2]
        ),
    )
    @mock.patch(
        "ax.modelbridge.base.ModelBridge.predict", autospec=True, return_value=None
    )
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def testGenWithDefaults(self, _, mock_gen):
        exp = get_experiment_for_value()
        exp.optimization_config = get_optimization_config_no_constraints()
        ss = get_search_space_for_range_value()
        modelbridge = ModelBridge(
            search_space=ss, model=Model(), transforms=[], experiment=exp
        )
        modelbridge.gen(1)
        mock_gen.assert_called_with(
            modelbridge,
            n=1,
            search_space=ss,
            fixed_features=ObservationFeatures(parameters={}),
            model_gen_options=None,
            optimization_config=OptimizationConfig(
                objective=Objective(metric=Metric("test_metric"), minimize=False),
                outcome_constraints=[],
            ),
            pending_observations={},
        )

    def test_transform_optimization_config(self) -> None:
        """
        The tested functionality is unused and is likely to be deprecated or
        removed, hence this test exists only to unbreak
        a failing codecov test. It is not an ideal test since we are using
        empty `fixed_features` and `transforms`.
        """
        ss = get_search_space_for_range_value()
        modelbridge = ModelBridge(search_space=ss, model=Model)

        fixed_features = ObservationFeatures(parameters={})
        optimization_config = get_optimization_config()
        new_cfg = modelbridge.transform_optimization_config(
            optimization_config, fixed_features
        )
        # In this case no transformations were applied, so config doesn't change
        self.assertEqual(optimization_config, new_cfg)
        # Even if no transforms were applied, method should return a new object
        self.assertFalse(optimization_config is new_cfg)

    @mock.patch(
        "ax.modelbridge.base.ModelBridge._gen",
        autospec=True,
        return_value=GenResults(
            observation_features=[get_observation1trans().features], weights=[2]
        ),
    )
    @mock.patch(
        "ax.modelbridge.base.ModelBridge.predict", autospec=True, return_value=None
    )
    # pyre-fixme[3]: Return type must be annotated.
    def test_gen_on_experiment_with_imm_ss_and_opt_conf(self, _, __):
        exp = get_experiment_for_value()
        exp._properties[Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF] = True
        exp.optimization_config = get_optimization_config_no_constraints()
        ss = get_search_space_for_range_value()
        modelbridge = ModelBridge(
            search_space=ss, model=Model(), transforms=[], experiment=exp
        )
        self.assertTrue(
            modelbridge._experiment_has_immutable_search_space_and_opt_config
        )
        gr = modelbridge.gen(1)
        self.assertIsNone(gr.optimization_config)
        self.assertIsNone(gr.search_space)

    @mock.patch(
        "ax.modelbridge.base.ModelBridge._gen",
        autospec=True,
        side_effect=[
            GenResults([get_observation1trans().features], [2]),
            GenResults([get_observation2trans().features], [2]),
            GenResults([get_observation2().features], weights=[2]),
        ],
    )
    @mock.patch("ax.modelbridge.base.ModelBridge._update", autospec=True)
    # pyre-fixme[3]: Return type must be annotated.
    def test_update(self, _mock_update, _mock_gen):
        exp = get_experiment_for_value()
        exp.optimization_config = get_optimization_config_no_constraints()
        ss = get_search_space_for_range_values(min=0, max=1000)
        exp.search_space = ss
        modelbridge = ModelBridge(
            search_space=ss, model=Model(), transforms=[Log], experiment=exp
        )
        exp.new_trial(generator_run=modelbridge.gen(1))
        modelbridge._set_training_data(
            observations_from_data(
                data=Data(
                    pd.DataFrame(
                        [
                            {
                                "arm_name": "0_0",
                                "metric_name": "m1",
                                "mean": 3.0,
                                "sem": 1.0,
                            }
                        ]
                    )
                ),
                experiment=exp,
            ),
            ss,
        )
        exp.new_trial(generator_run=modelbridge.gen(1))
        modelbridge.update(
            new_data=Data(
                pd.DataFrame(
                    [{"arm_name": "1_0", "metric_name": "m1", "mean": 5.0, "sem": 0.0}]
                )
            ),
            experiment=exp,
        )
        exp.new_trial(generator_run=modelbridge.gen(1))
        # Trying to update with unrecognised metric should error.
        with self.assertRaisesRegex(ValueError, "Unrecognised metric"):
            modelbridge.update(
                new_data=Data(
                    pd.DataFrame(
                        [
                            {
                                "arm_name": "1_0",
                                "metric_name": "m2",
                                "mean": 5.0,
                                "sem": 0.0,
                            }
                        ]
                    )
                ),
                experiment=exp,
            )


class testClampObservationFeatures(TestCase):
    def testClampObservationFeaturesNearBounds(self) -> None:
        cases = [
            (
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 2, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 2, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 0.0, "x": 2, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 0.5, "x": 2, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 100.0, "x": 2, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 5.5, "x": 2, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 0, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 1, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 11, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 10, "y": "foo", "z": True}
                ),
            ),
        ]
        search_space = get_experiment().search_space
        for obs_ft, expected_obs_ft in cases:
            actual_obs_ft = clamp_observation_features([obs_ft], search_space)
            self.assertEqual(actual_obs_ft[0], expected_obs_ft)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import dataclass, field
from typing import Any, Dict, List, Type

from ax.core.metric import Metric

from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    ObjectiveThreshold,
    OptimizationConfig,
)
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.runner import Runner
from ax.core.search_space import SearchSpace
from ax.core.types import ComparisonOp
from ax.metrics.botorch_test_problem import BotorchTestProblemMetric
from ax.runners.botorch_test_problem import BotorchTestProblemRunner
from ax.utils.common.base import Base
from botorch.test_functions.base import BaseTestProblem
from botorch.test_functions.multi_objective import MultiObjectiveTestProblem
from botorch.test_functions.synthetic import SyntheticTestFunction

# NOTE: Do not add `from __future__ import annotatations` to this file. Adding
# `annotations` postpones evaluation of types and will break FBLearner's usage of
# `BenchmarkProblem` as return type annotation, used for serialization and rendering
# in the UI.


@dataclass(frozen=True)
class BenchmarkProblem(Base):
    """Benchmark problem, represented in terms of Ax search space, optimization
    config, and runner.
    """

    name: str
    search_space: SearchSpace
    optimization_config: OptimizationConfig
    runner: Runner
    num_trials: int
    tracking_metrics: List[Metric] = field(default_factory=list)

    @classmethod
    def from_botorch(
        cls,
        test_problem_class: Type[BaseTestProblem],
        test_problem_kwargs: Dict[str, Any],
        num_trials: int,
    ) -> "BenchmarkProblem":
        """Create a BenchmarkProblem from a BoTorch BaseTestProblem using specialized
        Metrics and Runners. The test problem's result will be computed on the Runner
        and retrieved by the Metric.
        """

        # pyre-fixme [45]: Invalid class instantiation
        test_problem = test_problem_class(**test_problem_kwargs)

        search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    name=f"x{i}",
                    parameter_type=ParameterType.FLOAT,
                    lower=test_problem._bounds[i][0],
                    upper=test_problem._bounds[i][1],
                )
                for i in range(test_problem.dim)
            ]
        )

        optimization_config = OptimizationConfig(
            objective=Objective(
                metric=BotorchTestProblemMetric(
                    name=f"{test_problem.__class__.__name__}",
                    noise_sd=(test_problem.noise_std or 0),
                ),
                minimize=True,
            )
        )

        return cls(
            name=f"{test_problem.__class__.__name__}",
            search_space=search_space,
            optimization_config=optimization_config,
            runner=BotorchTestProblemRunner(
                test_problem_class=test_problem_class,
                test_problem_kwargs=test_problem_kwargs,
            ),
            num_trials=num_trials,
        )


@dataclass(frozen=True, init=False)
class SingleObjectiveBenchmarkProblem(BenchmarkProblem):
    """The most basic BenchmarkProblem, with a single objective and a known optimal
    value.
    """

    optimal_value: float = field()

    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, optimal_value: float, **kwargs) -> None:
        super().__init__(**kwargs)

        object.__setattr__(self, "optimal_value", optimal_value)

    @classmethod
    def from_botorch_synthetic(
        cls,
        test_problem_class: Type[SyntheticTestFunction],
        test_problem_kwargs: Dict[str, Any],
        num_trials: int,
    ) -> "SingleObjectiveBenchmarkProblem":
        """Create a BenchmarkProblem from a BoTorch BaseTestProblem using specialized
        Metrics and Runners. The test problem's result will be computed on the Runner
        and retrieved by the Metric.
        """

        # pyre-fixme [45]: Invalid class instantiation
        test_problem = test_problem_class(**test_problem_kwargs)

        problem = BenchmarkProblem.from_botorch(
            test_problem_class=test_problem_class,
            test_problem_kwargs=test_problem_kwargs,
            num_trials=num_trials,
        )

        return cls(
            name=f"{test_problem.__class__.__name__}",
            search_space=problem.search_space,
            optimization_config=problem.optimization_config,
            runner=problem.runner,
            num_trials=num_trials,
            optimal_value=test_problem.optimal_value,
        )


@dataclass(frozen=True, init=False)
class MultiObjectiveBenchmarkProblem(BenchmarkProblem):
    """A BenchmarkProblem support multiple objectives. Rather than knowing each
    objective's optimal value we track a known maximum hypervolume computed from a
    given reference point.
    """

    maximum_hypervolume: float = field()
    reference_point: List[float] = field()

    def __init__(
        self,
        maximum_hypervolume: float,
        reference_point: List[float],
        # pyre-fixme[2]: Parameter must be annotated.
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        object.__setattr__(self, "maximum_hypervolume", maximum_hypervolume)
        object.__setattr__(self, "reference_point", reference_point)

    @classmethod
    def from_botorch_multi_objective(
        cls,
        test_problem_class: Type[MultiObjectiveTestProblem],
        test_problem_kwargs: Dict[str, Any],
        num_trials: int,
    ) -> "MultiObjectiveBenchmarkProblem":
        """Create a BenchmarkProblem from a BoTorch BaseTestProblem using specialized
        Metrics and Runners. The test problem's result will be computed on the Runner
        once per trial and each Metric will retrieve its own result by index.
        """

        # pyre-fixme [45]: Invalid class instantiation
        test_problem = test_problem_class(**test_problem_kwargs)

        problem = BenchmarkProblem.from_botorch(
            test_problem_class=test_problem_class,
            test_problem_kwargs=test_problem_kwargs,
            num_trials=num_trials,
        )

        metrics = [
            BotorchTestProblemMetric(
                name=f"{test_problem.__class__.__name__}_{i}",
                noise_sd=(test_problem.noise_std or 0),
                index=i,
            )
            for i in range(test_problem.num_objectives)
        ]
        optimization_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(
                objectives=[
                    Objective(
                        metric=metric,
                        minimize=True,
                    )
                    for metric in metrics
                ]
            ),
            objective_thresholds=[
                ObjectiveThreshold(
                    metric=metrics[i],
                    # pyre-fixme[6]: For 2nd param expected `float` but got `Tensor`.
                    bound=test_problem.ref_point[i],
                    relative=False,
                    op=ComparisonOp.LEQ,
                )
                for i in range(test_problem.num_objectives)
            ],
        )

        return cls(
            name=f"{test_problem.__class__.__name__}",
            search_space=problem.search_space,
            optimization_config=optimization_config,
            runner=problem.runner,
            num_trials=num_trials,
            maximum_hypervolume=test_problem.max_hv,
            reference_point=test_problem._ref_point,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import dataclasses
import warnings
from abc import ABC
from contextlib import ExitStack
from itertools import product
from logging import Logger
from math import sqrt
from typing import Dict, Type
from unittest import mock

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import AxError
from ax.models.torch.fully_bayesian import (
    FullyBayesianBotorchModel,
    FullyBayesianMOOBotorchModel,
    matern_kernel,
    rbf_kernel,
    single_task_pyro_model,
)
from ax.models.torch_base import TorchOptConfig
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from ax.utils.common.testutils import TestCase
from ax.utils.testing.torch_stubs import get_torch_test_data
from botorch.acquisition.utils import get_infeasible_cost
from botorch.models import ModelListGP
from botorch.models.gp_regression import MIN_INFERRED_NOISE_LEVEL
from botorch.models.transforms.input import Warp
from botorch.optim.optimize import optimize_acqf
from botorch.utils import get_objective_weights_transform
from botorch.utils.datasets import FixedNoiseDataset
from gpytorch.likelihoods import _GaussianLikelihoodBase

from pyro.infer.mcmc import MCMC, NUTS  # @manual=//third-party-source/native/pyro:pyro

RUN_INFERENCE_PATH = "ax.models.torch.fully_bayesian.run_inference"
NUTS_PATH = "pyro.infer.mcmc.NUTS"
MCMC_PATH = "pyro.infer.mcmc.MCMC"

logger: Logger = get_logger(__name__)


def _get_dummy_mcmc_samples(
    num_samples: int,
    num_outputs: int,
    dtype: torch.dtype,
    device: torch.device,
    perturb_sd: float = 1e-6,
) -> Dict[str, torch.Tensor]:
    tkwargs = {"dtype": dtype, "device": device}
    dummy_sample_list = []
    for i in range(num_outputs):
        dummy_samples = {
            # use real MAP values with tiny perturbations
            # so that the generation code below has feasible in-sample
            # points
            # pyre-fixme[6]: For 2nd param expected `Optional[dtype]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Union[None, str, device]` but
            #  got `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `bool` but got `Union[device,
            #  dtype]`.
            "lengthscale": (i + 1) * torch.tensor([[1 / 3, 1 / 3, 1 / 3]], **tkwargs)
            # pyre-fixme[6]: For 4th param expected `Optional[Sequence[Union[None,
            #  ellipsis, str]]]` but got `Union[device, dtype]`.
            # pyre-fixme[6]: For 4th param expected `Optional[bool]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 4th param expected `Optional[Generator]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 4th param expected `Optional[dtype]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 4th param expected `Optional[layout]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 4th param expected `Union[None, str, device]` but
            #  got `Union[device, dtype]`.
            + perturb_sd * torch.randn(num_samples, 1, 3, **tkwargs),
            # pyre-fixme[6]: For 2nd param expected `Optional[dtype]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Union[None, str, device]` but
            #  got `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `bool` but got `Union[device,
            #  dtype]`.
            "outputscale": (i + 1) * torch.tensor(2.3436, **tkwargs)
            # pyre-fixme[6]: For 2nd param expected `Optional[Sequence[Union[None,
            #  ellipsis, str]]]` but got `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[bool]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[Generator]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[dtype]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[layout]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Union[None, str, device]` but
            #  got `Union[device, dtype]`.
            + perturb_sd * torch.randn(num_samples, **tkwargs),
            # pyre-fixme[6]: For 2nd param expected `Optional[dtype]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Union[None, str, device]` but
            #  got `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `bool` but got `Union[device,
            #  dtype]`.
            "mean": (i + 1) * torch.tensor([3.5000], **tkwargs)
            # pyre-fixme[6]: For 2nd param expected `Optional[Sequence[Union[None,
            #  ellipsis, str]]]` but got `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[bool]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[Generator]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[dtype]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Optional[layout]` but got
            #  `Union[device, dtype]`.
            # pyre-fixme[6]: For 2nd param expected `Union[None, str, device]` but
            #  got `Union[device, dtype]`.
            + perturb_sd * torch.randn(num_samples, **tkwargs),
        }
        # pyre-fixme[6]: For 2nd param expected `Optional[dtype]` but got
        #  `Union[device, dtype]`.
        # pyre-fixme[6]: For 2nd param expected `Union[None, str, device]` but got
        #  `Union[device, dtype]`.
        # pyre-fixme[6]: For 2nd param expected `bool` but got `Union[device, dtype]`.
        dummy_samples["kernel_tausq"] = (i + 1) * torch.tensor(0.5, **tkwargs)
        dummy_samples["_kernel_inv_length_sq"] = (
            # pyre-fixme[6]: For 2nd param expected `Tensor` but got `float`.
            1.0
            # pyre-fixme[58]: `/` is not supported for operand types `float` and
            #  `Tensor`.
            / dummy_samples["lengthscale"].sqrt()
        )
        dummy_sample_list.append(dummy_samples)
    # pyre-fixme[7]: Expected `Dict[str, Tensor]` but got `List[typing.Any]`.
    return dummy_sample_list


def dummy_func(X: torch.Tensor) -> torch.Tensor:
    return X


class BaseFullyBayesianBotorchModelTest(ABC):
    model_cls: Type[FullyBayesianBotorchModel]

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_FullyBayesianBotorchModel(self, dtype=torch.float, cuda=False):
        # test deprecation warning
        warnings.resetwarnings()  # this is necessary for building in mode/opt
        warnings.simplefilter("always", append=True)
        with warnings.catch_warnings(record=True) as ws:
            self.model_cls(use_saas=True)
            # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
            #  attribute `assertTrue`.
            self.assertTrue(any(issubclass(w.category, DeprecationWarning) for w in ws))
            msg = "Passing `use_saas` is no longer supported"
            self.assertTrue(any(msg in str(w.message) for w in ws))
        Xs1, Ys1, Yvars1, bounds, tfs, fns, mns = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        Yvars_inferred_noise = [
            torch.full_like(Yvars1[0], float("nan")),
            torch.full_like(Yvars2[0], float("nan")),
        ]
        # make input different for each output
        Xs2_diff = [Xs2[0] + 0.1]
        Xs = Xs1 + Xs2_diff
        Ys = Ys1 + Ys2

        options = product([True, False], [True, False], ["matern", "rbf"])
        for inferred_noise, use_input_warping, gp_kernel in options:
            Yvars = Yvars_inferred_noise if inferred_noise else Yvars1 + Yvars2
            model = self.model_cls(
                use_input_warping=use_input_warping,
                thinning=1,
                num_samples=4,
                disable_progbar=True,
                max_tree_depth=1,
                gp_kernel=gp_kernel,
                verbose=True,
            )
            if use_input_warping:
                self.assertTrue(model.use_input_warping)
            # Test ModelListGP
            # make training data different for each output
            tkwargs = {"dtype": dtype, "device": Xs1[0].device}
            dummy_samples_list = _get_dummy_mcmc_samples(
                num_samples=4, num_outputs=2, **tkwargs
            )
            for dummy_samples in dummy_samples_list:
                if use_input_warping:
                    # pyre-fixme[16]: `str` has no attribute `__setitem__`.
                    dummy_samples["c0"] = (
                        torch.rand(4, 1, Xs1[0].shape[-1], **tkwargs) * 0.5 + 0.1
                    )
                    dummy_samples["c1"] = (
                        torch.rand(4, 1, Xs1[0].shape[-1], **tkwargs) * 0.5 + 0.1
                    )
                if inferred_noise:
                    dummy_samples["noise"] = torch.rand(4, 1, **tkwargs).clamp_min(
                        MIN_INFERRED_NOISE_LEVEL
                    )

            with mock.patch(
                RUN_INFERENCE_PATH,
                side_effect=dummy_samples_list,
            ) as _mock_fit_model:
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs, Ys, Yvars)
                    ],
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                    metric_names=["y1", "y2"],
                )
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertEqual`.
                self.assertEqual(_mock_fit_model.call_count, 2)
                for i, call in enumerate(_mock_fit_model.call_args_list):
                    _, ckwargs = call
                    X = Xs[i]
                    Y = Ys[i]
                    Yvar = Yvars[i]
                    # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                    #  attribute `assertIs`.
                    self.assertIs(ckwargs["pyro_model"], single_task_pyro_model)

                    self.assertTrue(torch.equal(ckwargs["X"], X))
                    self.assertTrue(torch.equal(ckwargs["Y"], Y))
                    if inferred_noise:
                        self.assertTrue(torch.isnan(ckwargs["Yvar"]).all())
                    else:
                        self.assertTrue(torch.equal(ckwargs["Yvar"], Yvar))
                    self.assertEqual(ckwargs["num_samples"], 4)
                    self.assertEqual(ckwargs["warmup_steps"], 512)
                    self.assertEqual(ckwargs["max_tree_depth"], 1)
                    self.assertTrue(ckwargs["disable_progbar"])
                    self.assertEqual(ckwargs["use_input_warping"], use_input_warping)
                    self.assertEqual(ckwargs["gp_kernel"], gp_kernel)
                    self.assertTrue(ckwargs["verbose"])

                    # Check attributes
                    self.assertTrue(torch.equal(model.Xs[i], Xs[i]))
                    self.assertEqual(model.dtype, Xs[i].dtype)
                    self.assertEqual(model.device, Xs[i].device)
                    # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                    #  attribute `assertIsInstance`.
                    self.assertIsInstance(model.model, ModelListGP)

                    # Check fitting
                    # Note each model in the model list is a batched model, where
                    # the batch dim corresponds to the MCMC samples
                    # pyre-fixme[16]: Optional type has no attribute `models`.
                    model_list = model.model.models
                    # Put model in `eval` mode to transform the train inputs.
                    m = model_list[i].eval()
                    # check mcmc samples
                    # pyre-fixme[6]: For 1st param expected `str` but got `int`.
                    dummy_samples = dummy_samples_list[i]
                    expected_train_inputs = Xs[i].expand(4, *Xs[i].shape)
                    if use_input_warping:
                        # train inputs should be warped inputs
                        expected_train_inputs = m.input_transform(expected_train_inputs)
                    self.assertTrue(
                        torch.equal(
                            m.train_inputs[0],
                            expected_train_inputs,
                        )
                    )
                    self.assertTrue(
                        torch.equal(
                            m.train_targets,
                            Ys[i].view(1, -1).expand(4, Ys[i].numel()),
                        )
                    )
                    expected_noise = (
                        # pyre-fixme[6]: For 1st param expected `Union[None,
                        #  List[typing.Any], int, slice, Tensor,
                        #  typing.Tuple[typing.Any, ...]]` but got `str`.
                        dummy_samples["noise"].view(m.likelihood.noise.shape)
                        if inferred_noise
                        else Yvars[i].view(1, -1).expand(4, Yvars[i].numel())
                    )
                    self.assertTrue(
                        torch.allclose(
                            m.likelihood.noise.detach(),
                            expected_noise,
                        )
                    )
                    self.assertIsInstance(m.likelihood, _GaussianLikelihoodBase)
                    self.assertTrue(
                        torch.allclose(
                            m.covar_module.base_kernel.lengthscale.detach(),
                            # pyre-fixme[6]: For 1st param expected `Union[None,
                            #  List[typing.Any], int, slice, Tensor,
                            #  typing.Tuple[typing.Any, ...]]` but got `str`.
                            dummy_samples["lengthscale"].view(
                                m.covar_module.base_kernel.lengthscale.shape
                            ),
                            rtol=1e-4,
                            atol=1e-6,
                        )
                    )
                    self.assertTrue(
                        torch.allclose(
                            m.covar_module.outputscale.detach(),
                            # pyre-fixme[6]: For 1st param expected `Union[None,
                            #  List[typing.Any], int, slice, Tensor,
                            #  typing.Tuple[typing.Any, ...]]` but got `str`.
                            dummy_samples["outputscale"].view(
                                m.covar_module.outputscale.shape
                            ),
                        )
                    )
                    self.assertTrue(
                        torch.allclose(
                            m.mean_module.constant.detach(),
                            # pyre-fixme[6]: For 1st param expected `Union[None,
                            #  List[typing.Any], int, slice, Tensor,
                            #  typing.Tuple[typing.Any, ...]]` but got `str`.
                            dummy_samples["mean"].view(m.mean_module.constant.shape),
                        )
                    )
                    if use_input_warping:
                        self.assertTrue(hasattr(m, "input_transform"))
                        self.assertIsInstance(m.input_transform, Warp)
                        self.assertTrue(
                            torch.equal(
                                m.input_transform.concentration0,
                                # pyre-fixme[6]: For 1st param expected `str`
                                #  but got `int`.
                                # pyre-fixme[6]: For 1st param expected
                                #  `Union[None, List[typing.Any], int, slice,
                                #  Tensor, typing.Tuple[typing.Any, ...]]` but got
                                #  `str`.
                                dummy_samples_list[i]["c0"],
                            )
                        )
                        self.assertTrue(
                            torch.equal(
                                m.input_transform.concentration1,
                                # pyre-fixme[6]: For 1st param expected `str`
                                #  but got `int`.
                                # pyre-fixme[6]: For 1st param expected
                                #  `Union[None, List[typing.Any], int, slice,
                                #  Tensor, typing.Tuple[typing.Any, ...]]` but got
                                #  `str`.
                                dummy_samples_list[i]["c1"],
                            )
                        )
                    else:
                        # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest`
                        #  has no attribute `assertFalse`.
                        self.assertFalse(hasattr(m, "input_transform"))
            # test that multi-task is not implemented
            (
                Xs_mt,
                Ys_mt,
                Yvars_mt,
                bounds_mt,
                tfs_mt,
                fns_mt,
                mns_mt,
            ) = get_torch_test_data(
                dtype=dtype, cuda=cuda, constant_noise=True, task_features=[2]
            )
            with mock.patch(
                RUN_INFERENCE_PATH,
                side_effect=dummy_samples_list,
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertRaises`.
            ) as _mock_fit_model, self.assertRaises(NotImplementedError):
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs_mt, Ys_mt, Yvars_mt)
                    ],
                    metric_names=mns_mt,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns_mt,
                        bounds=bounds_mt,
                        task_features=tfs_mt,
                    ),
                )
            with mock.patch(
                RUN_INFERENCE_PATH,
                side_effect=dummy_samples_list,
            ) as _mock_fit_model, self.assertRaises(NotImplementedError):
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars1 + Yvars2)
                    ],
                    metric_names=mns,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns,
                        bounds=bounds,
                        fidelity_features=[0],
                    ),
                )
            # fit model with same inputs (otherwise X_observed will be None)
            model = self.model_cls(
                use_input_warping=use_input_warping,
                thinning=1,
                num_samples=4,
                disable_progbar=True,
                max_tree_depth=1,
                gp_kernel=gp_kernel,
            )
            Yvars = Yvars1 + Yvars2
            dummy_samples_list = _get_dummy_mcmc_samples(
                num_samples=4, num_outputs=2, **tkwargs
            )
            with mock.patch(
                RUN_INFERENCE_PATH,
                side_effect=dummy_samples_list,
            ) as _mock_fit_model:
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars)
                    ],
                    metric_names=mns,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                )

            # Check the hyperparameters and shapes
            self.assertEqual(len(model.model.models), 2)
            m1, m2 = model.model.models[0], model.model.models[1]
            # Mean
            self.assertEqual(m1.mean_module.constant.shape, (4,))
            self.assertFalse(
                torch.isclose(m1.mean_module.constant, m2.mean_module.constant).any()
            )
            # Outputscales
            self.assertEqual(m1.covar_module.outputscale.shape, (4,))
            self.assertFalse(
                torch.isclose(
                    m1.covar_module.outputscale, m2.covar_module.outputscale
                ).any()
            )
            # Lengthscales
            self.assertEqual(m1.covar_module.base_kernel.lengthscale.shape, (4, 1, 3))
            self.assertFalse(
                torch.isclose(
                    m1.covar_module.base_kernel.lengthscale,
                    m2.covar_module.base_kernel.lengthscale,
                ).any()
            )

            # Check infeasible cost can be computed on the model
            device = torch.device("cuda") if cuda else torch.device("cpu")
            objective_weights = torch.tensor([1.0, 0.0], dtype=dtype, device=device)
            objective_transform = get_objective_weights_transform(objective_weights)
            infeasible_cost = (
                get_infeasible_cost(
                    X=Xs1[0],
                    # pyre-fixme[6]: For 2nd param expected `Model` but got
                    #  `Optional[Model]`.
                    model=model.model,
                    objective=objective_transform,
                )
                .detach()
                .clone()
            )
            expected_infeasible_cost = -1 * torch.min(
                # pyre-fixme[20]: Argument `1` expected.
                objective_transform(
                    # pyre-fixme[16]: Optional type has no attribute `posterior`.
                    model.model.posterior(Xs1[0]).mean
                    - 6 * model.model.posterior(Xs1[0]).variance.sqrt()
                ).min(),
                torch.tensor(0.0, dtype=dtype, device=device),
            )
            self.assertTrue(
                torch.abs(infeasible_cost - expected_infeasible_cost) < 1e-5
            )

            # Check prediction
            X = torch.tensor([[6.0, 7.0, 8.0]], **tkwargs)
            f_mean, f_cov = model.predict(X)
            self.assertTrue(f_mean.shape == torch.Size([1, 2]))
            self.assertTrue(f_cov.shape == torch.Size([1, 2, 2]))

            # Check generation
            objective_weights = torch.tensor(
                [1.0, 0.0]
                if self.model_cls is FullyBayesianBotorchModel
                else [1.0, 1.0],
                **tkwargs,
            )
            outcome_constraints = (
                torch.tensor([[0.0, 1.0]], **tkwargs),
                torch.tensor([[5.0]], **tkwargs),
            )
            linear_constraints = (
                torch.tensor([[0.0, 1.0, 1.0]]),
                torch.tensor([[100.0]]),
            )
            fixed_features = None
            pending_observations = [
                torch.tensor([[1.0, 3.0, 4.0]], **tkwargs),
                torch.tensor([[2.0, 6.0, 8.0]], **tkwargs),
            ]
            n = 3

            X_dummy = torch.tensor([[[1.0, 2.0, 3.0]]], **tkwargs)
            acqfv_dummy = torch.tensor([[[1.0, 2.0, 3.0]]], **tkwargs)
            model_gen_options = {
                Keys.OPTIMIZER_KWARGS: {"maxiter": 1},
                Keys.ACQF_KWARGS: {"mc_samples": 3},
            }
            search_space_digest = SearchSpaceDigest(
                feature_names=fns,
                bounds=bounds,
            )
            torch_opt_config = TorchOptConfig(
                objective_weights=objective_weights,
                objective_thresholds=torch.zeros(2, **tkwargs)
                if self.model_cls is FullyBayesianMOOBotorchModel
                else None,
                outcome_constraints=outcome_constraints,
                linear_constraints=linear_constraints,
                fixed_features=fixed_features,
                pending_observations=pending_observations,
                # pyre-fixme[6]: For 7th param expected `Dict[str, Union[None,
                #  Dict[str, typing.Any], OptimizationConfig, AcquisitionFunction,
                #  float, int, str]]` but got `Dict[Keys, Dict[str, int]]`.
                model_gen_options=model_gen_options,
                rounding_func=dummy_func,
                is_moo=self.model_cls is FullyBayesianMOOBotorchModel,
            )
            # test sequential optimize with constraints
            with mock.patch(
                "ax.models.torch.botorch_defaults.optimize_acqf",
                return_value=(X_dummy, acqfv_dummy),
            ) as _:
                # Xgen, wgen, gen_metadata, cand_metadata = model.gen(
                gen_results = model.gen(
                    n=n,
                    search_space_digest=search_space_digest,
                    torch_opt_config=torch_opt_config,
                )
                # note: gen() always returns CPU tensors
                self.assertTrue(torch.equal(gen_results.points, X_dummy.cpu()))
                self.assertTrue(
                    torch.equal(gen_results.weights, torch.ones(n, dtype=dtype))
                )

            # actually test optimization for 1 step without constraints
            with mock.patch(
                "ax.models.torch.botorch_defaults.optimize_acqf",
                wraps=optimize_acqf,
                return_value=(X_dummy, acqfv_dummy),
            ) as _:
                gen_results = model.gen(
                    n=n,
                    search_space_digest=search_space_digest,
                    torch_opt_config=dataclasses.replace(
                        torch_opt_config, linear_constraints=None
                    ),
                )
                # note: gen() always returns CPU tensors
                self.assertTrue(torch.equal(gen_results.points, X_dummy.cpu()))
                self.assertTrue(
                    torch.equal(gen_results.weights, torch.ones(n, dtype=dtype))
                )

            # Check best point selection
            if self.model_cls is FullyBayesianMOOBotorchModel:
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertRaisesRegex`.
                with self.assertRaisesRegex(NotImplementedError, "Best observed"):
                    model.best_point(
                        search_space_digest=search_space_digest,
                        torch_opt_config=torch_opt_config,
                    )
            else:
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertIsNotNone`.
                self.assertIsNotNone(
                    model.best_point(
                        search_space_digest=search_space_digest,
                        torch_opt_config=torch_opt_config,
                    )
                )
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertIsNone`.
                self.assertIsNone(
                    model.best_point(
                        search_space_digest=search_space_digest,
                        torch_opt_config=dataclasses.replace(
                            torch_opt_config, fixed_features={0: 100.0}
                        ),
                    )
                )

            # Test cross-validation
            mean, variance = model.cross_validate(
                datasets=[
                    FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                    for X, Y, Yvar in zip(Xs1 + Xs2, Ys, Yvars)
                ],
                X_test=torch.tensor(
                    [[1.2, 3.2, 4.2], [2.4, 5.2, 3.2]], dtype=dtype, device=device
                ),
            )
            self.assertTrue(mean.shape == torch.Size([2, 2]))
            self.assertTrue(variance.shape == torch.Size([2, 2, 2]))

            # Test cross-validation with refit_on_cv
            model.refit_on_cv = True
            with mock.patch(
                RUN_INFERENCE_PATH,
                side_effect=dummy_samples_list,
            ) as _mock_fit_model:
                mean, variance = model.cross_validate(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys, Yvars)
                    ],
                    X_test=torch.tensor(
                        [[1.2, 3.2, 4.2], [2.4, 5.2, 3.2]],
                        dtype=dtype,
                        device=device,
                    ),
                )
                self.assertTrue(mean.shape == torch.Size([2, 2]))
                self.assertTrue(variance.shape == torch.Size([2, 2, 2]))

            # Test update
            model.refit_on_update = False
            model.update(
                datasets=[
                    FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                    for X, Y, Yvar in zip(Xs2 + Xs2, Ys2 + Ys2, Yvars2 + Yvars2)
                ]
            )

            # Test feature_importances
            importances = model.feature_importances()
            self.assertEqual(importances.shape, torch.Size([2, 1, 3]))

            # When calling update directly, the data is completely overwritten.
            self.assertTrue(torch.equal(model.Xs[0], Xs2[0]))
            self.assertTrue(torch.equal(model.Xs[1], Xs2[0]))
            self.assertTrue(torch.equal(model.Ys[0], Ys2[0]))
            self.assertTrue(torch.equal(model.Yvars[0], Yvars2[0]))

            model.refit_on_update = True
            with mock.patch(
                RUN_INFERENCE_PATH, side_effect=dummy_samples_list
            ) as _mock_fit_model:
                model.update(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs2 + Xs2, Ys2 + Ys2, Yvars2 + Yvars2)
                    ]
                )

            # test unfit model CV, update, and feature_importances
            unfit_model = self.model_cls()
            with self.assertRaises(RuntimeError):
                unfit_model.cross_validate(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars1 + Yvars2)
                    ],
                    X_test=Xs1[0],
                )
            with self.assertRaises(RuntimeError):
                unfit_model.update(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars1 + Yvars2)
                    ]
                )
            with self.assertRaises(RuntimeError):
                unfit_model.feature_importances()

    def test_saasbo_sample(self) -> None:
        for use_input_warping, gp_kernel in product([False, True], ["rbf", "matern"]):
            with torch.random.fork_rng():
                torch.manual_seed(0)
                X = torch.randn(3, 2)
                Y = torch.randn(3, 1)
                Yvar = torch.randn(3, 1)
                kernel = NUTS(single_task_pyro_model, max_tree_depth=1)
                mcmc = MCMC(kernel, warmup_steps=0, num_samples=1)
                mcmc.run(
                    X,
                    Y,
                    Yvar,
                    use_input_warping=use_input_warping,
                    gp_kernel=gp_kernel,
                )
                samples = mcmc.get_samples()
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertTrue`.
                self.assertTrue("kernel_tausq" in samples)
                self.assertTrue("_kernel_inv_length_sq" in samples)
                self.assertTrue("lengthscale" not in samples)
                if use_input_warping:
                    # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                    #  attribute `assertIn`.
                    self.assertIn("c0", samples)
                    self.assertIn("c1", samples)
                else:
                    # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                    #  attribute `assertNotIn`.
                    self.assertNotIn("c0", samples)
                    self.assertNotIn("c1", samples)

    def test_gp_kernels(self) -> None:
        torch.manual_seed(0)
        X = torch.randn(3, 2)
        Y = torch.randn(3, 1)
        Yvar = torch.randn(3, 1)
        kernel = NUTS(single_task_pyro_model, max_tree_depth=1)
        # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no attribute
        #  `assertRaises`.
        with self.assertRaises(ValueError):
            mcmc = MCMC(kernel, warmup_steps=0, num_samples=1)
            mcmc.run(
                X,
                Y,
                Yvar,
                gp_kernel="some_kernel_we_dont_support",
            )

    def test_FullyBayesianBotorchModel_cuda(self) -> None:
        if torch.cuda.is_available():
            self.test_FullyBayesianBotorchModel(cuda=True)

    def test_FullyBayesianBotorchModel_double(self) -> None:
        self.test_FullyBayesianBotorchModel(dtype=torch.double)

    def test_FullyBayesianBotorchModel_double_cuda(self) -> None:
        if torch.cuda.is_available():
            self.test_FullyBayesianBotorchModel(dtype=torch.double, cuda=True)

    def test_FullyBayesianBotorchModelConstraints(self) -> None:
        Xs1, Ys1, Yvars1, bounds, tfs, fns, mns = get_torch_test_data(
            dtype=torch.float, cuda=False, constant_noise=True
        )
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=torch.float, cuda=False, constant_noise=True
        )
        # make infeasible
        Xs2[0] = -1 * Xs2[0]
        objective_weights = torch.tensor(
            [-1.0, 1.0], dtype=torch.float, device=torch.device("cpu")
        )
        n = 3
        model = self.model_cls(
            num_samples=4,
            thinning=1,
            disable_progbar=True,
            max_tree_depth=1,
        )
        dummy_samples = _get_dummy_mcmc_samples(
            num_samples=4, num_outputs=2, dtype=torch.float, device=Xs1[0].device
        )
        search_space_digest = SearchSpaceDigest(
            feature_names=fns,
            bounds=bounds,
            task_features=tfs,
        )
        with mock.patch(
            RUN_INFERENCE_PATH, side_effect=dummy_samples
        ) as _mock_fit_model:
            model.fit(
                datasets=[
                    FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                    for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars1 + Yvars2)
                ],
                metric_names=mns,
                search_space_digest=search_space_digest,
            )
            # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
            #  attribute `assertEqual`.
            self.assertEqual(_mock_fit_model.call_count, 2)

        # because there are no feasible points:
        # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no attribute
        #  `assertRaises`.
        with self.assertRaises(ValueError):
            model.gen(
                n,
                search_space_digest=search_space_digest,
                torch_opt_config=TorchOptConfig(objective_weights=objective_weights),
            )

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def test_FullyBayesianBotorchModelPyro(self, dtype=torch.double, cuda=False):
        Xs1, Ys1, raw_Yvars1, bounds, tfs, fns, mns = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        Xs2, Ys2, raw_Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        options = [(False, True, "rbf"), (True, False, "matern")]
        for inferred_noise, use_input_warping, gp_kernel in options:
            model = self.model_cls(
                num_samples=4,
                warmup_steps=0,
                thinning=1,
                use_input_warping=use_input_warping,
                disable_progbar=True,
                max_tree_depth=1,
                gp_kernel=gp_kernel,
                verbose=True,
            )
            if inferred_noise:
                Yvars1 = [torch.full_like(raw_Yvars1[0], float("nan"))]
                Yvars2 = [torch.full_like(raw_Yvars2[0], float("nan"))]
            else:
                Yvars1 = raw_Yvars1
                Yvars2 = raw_Yvars2

            dummy_samples = _get_dummy_mcmc_samples(
                num_samples=4,
                num_outputs=2,
                dtype=dtype,
                device=Xs1[0].device,
            )
            with ExitStack() as es:
                _mock_fit_model = es.enter_context(
                    mock.patch(RUN_INFERENCE_PATH, side_effect=dummy_samples)
                )
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars1 + Yvars2)
                    ],
                    metric_names=mns,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                )
                # check run_inference arguments
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertEqual`.
                self.assertEqual(_mock_fit_model.call_count, 2)
                _, ckwargs = _mock_fit_model.call_args
                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertIs`.
                self.assertIs(ckwargs["pyro_model"], single_task_pyro_model)

                # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest` has no
                #  attribute `assertTrue`.
                self.assertTrue(torch.equal(ckwargs["X"], Xs1[0]))
                self.assertTrue(torch.equal(ckwargs["Y"], Ys1[0]))
                if inferred_noise:
                    self.assertTrue(torch.isnan(ckwargs["Yvar"]).all())
                else:
                    self.assertTrue(torch.equal(ckwargs["Yvar"], Yvars1[0]))
                self.assertEqual(ckwargs["num_samples"], 4)
                self.assertEqual(ckwargs["warmup_steps"], 0)
                self.assertEqual(ckwargs["max_tree_depth"], 1)
                self.assertTrue(ckwargs["disable_progbar"])
                self.assertEqual(ckwargs["use_input_warping"], use_input_warping)
                self.assertEqual(ckwargs["gp_kernel"], gp_kernel)
                self.assertTrue(ckwargs["verbose"])

            with ExitStack() as es:
                _mock_mcmc = es.enter_context(mock.patch(MCMC_PATH))
                _mock_mcmc.return_value.get_samples.side_effect = dummy_samples
                _mock_nuts = es.enter_context(mock.patch(NUTS_PATH))
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars1 + Yvars2)
                    ],
                    metric_names=mns,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                )
                # check MCMC.__init__ arguments
                self.assertEqual(_mock_mcmc.call_count, 2)
                _, ckwargs = _mock_mcmc.call_args
                self.assertEqual(ckwargs["num_samples"], 4)
                self.assertEqual(ckwargs["warmup_steps"], 0)
                self.assertTrue(ckwargs["disable_progbar"])
                # check NUTS.__init__ arguments
                _mock_nuts.assert_called_with(
                    single_task_pyro_model,
                    jit_compile=False,
                    full_mass=True,
                    ignore_jit_warnings=True,
                    max_tree_depth=1,
                )
            # now actually run pyro
            if not use_input_warping:
                # input warping is quite slow, so we omit it for
                # testing purposes
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1 + Xs2, Ys1 + Ys2, Yvars1 + Yvars2)
                    ],
                    metric_names=mns,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                )

                for m, X, Y, Yvar in zip(
                    # pyre-fixme[16]: Optional type has no attribute `models`.
                    model.model.models,
                    Xs1 + Xs2,
                    Ys1 + Ys2,
                    Yvars1 + Yvars2,
                ):
                    self.assertTrue(
                        torch.equal(
                            m.train_inputs[0],
                            X.expand(4, *X.shape),
                        )
                    )
                    self.assertTrue(
                        torch.equal(
                            m.train_targets,
                            Y.view(1, -1).expand(4, Y.numel()),
                        )
                    )
                    # check shapes of sampled parameters
                    if not inferred_noise:
                        self.assertTrue(
                            torch.allclose(
                                m.likelihood.noise.detach(),
                                Yvar.view(1, -1).expand(4, Yvar.numel()),
                            )
                        )
                    else:
                        self.assertEqual(m.likelihood.noise.shape, torch.Size([4, 1]))

                    self.assertEqual(
                        m.covar_module.base_kernel.lengthscale.shape,
                        torch.Size([4, 1, X.shape[-1]]),
                    )
                    self.assertEqual(m.covar_module.outputscale.shape, torch.Size([4]))
                    self.assertEqual(
                        m.mean_module.constant.shape,
                        torch.Size([4]),
                    )
                    if use_input_warping:
                        self.assertTrue(hasattr(m, "input_transform"))
                        # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest`
                        #  has no attribute `assertIsInstance`.
                        self.assertIsInstance(m.input_transform, Warp)
                        self.assertEqual(
                            m.input_transform.concentration0.shape,
                            torch.Size([4, 1, 3]),
                        )
                        self.assertEqual(
                            m.input_transform.concentration1.shape,
                            torch.Size([4, 1, 3]),
                        )
                    else:
                        # pyre-fixme[16]: `BaseFullyBayesianBotorchModelTest`
                        #  has no attribute `assertFalse`.
                        self.assertFalse(hasattr(m, "input_transform"))

    def test_FullyBayesianBotorchModelPyro_float(self) -> None:
        self.test_FullyBayesianBotorchModelPyro(dtype=torch.float, cuda=False)

    def test_FullyBayesianBotorchModelPyro_cuda_double(self) -> None:
        if torch.cuda.is_available():
            self.test_FullyBayesianBotorchModelPyro(dtype=torch.double, cuda=True)

    def test_FullyBayesianBotorchModelPyro_cuda_float(self) -> None:
        if torch.cuda.is_available():
            self.test_FullyBayesianBotorchModelPyro(dtype=torch.float, cuda=True)


class SingleObjectiveFullyBayesianBotorchModelTest(
    TestCase, BaseFullyBayesianBotorchModelTest
):
    model_cls = FullyBayesianBotorchModel

    def test_FullyBayesianBotorchModelOneOutcome(self) -> None:
        Xs1, Ys1, Yvars1, bounds, tfs, fns, mns = get_torch_test_data(
            dtype=torch.float, cuda=False, constant_noise=True
        )
        for use_input_warping, gp_kernel in product([True, False], ["rbf", "matern"]):
            model = self.model_cls(
                use_input_warping=use_input_warping,
                num_samples=4,
                thinning=1,
                disable_progbar=True,
                max_tree_depth=1,
                gp_kernel=gp_kernel,
            )
            dummy_samples = _get_dummy_mcmc_samples(
                num_samples=4,
                num_outputs=1,
                dtype=torch.float,
                device=Xs1[0].device,
            )
            with mock.patch(
                RUN_INFERENCE_PATH, side_effect=dummy_samples
            ) as _mock_fit_model:
                model.fit(
                    datasets=[
                        FixedNoiseDataset(X=X, Y=Y, Yvar=Yvar)
                        for X, Y, Yvar in zip(Xs1, Ys1, Yvars1)
                    ],
                    metric_names=[mns[0]],
                    search_space_digest=SearchSpaceDigest(
                        feature_names=fns,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                )
                _mock_fit_model.assert_called_once()
            X = torch.rand(2, 3, dtype=torch.float)
            f_mean, f_cov = model.predict(X)
            self.assertTrue(f_mean.shape == torch.Size([2, 1]))
            self.assertTrue(f_cov.shape == torch.Size([2, 1, 1]))
            # pyre-fixme[16]: Optional type has no attribute `models`.
            model_list = model.model.models
            self.assertTrue(len(model_list) == 1)
            if use_input_warping:
                self.assertTrue(hasattr(model_list[0], "input_transform"))
                self.assertIsInstance(model_list[0].input_transform, Warp)
            else:
                self.assertFalse(hasattr(model_list[0], "input_transform"))


class FullyBayesianMOOBotorchModelTest(TestCase, BaseFullyBayesianBotorchModelTest):
    # pyre-fixme[15]: `model_cls` overrides attribute defined in
    #  `BaseFullyBayesianBotorchModelTest` inconsistently.
    model_cls = FullyBayesianMOOBotorchModel


class TestKernels(TestCase):
    def test_matern_kernel(self) -> None:
        a = torch.tensor([4, 2, 8], dtype=torch.float).view(3, 1)
        b = torch.tensor([0, 2], dtype=torch.float).view(2, 1)
        lengthscale = 2
        # test matern 1/2
        # pyre-fixme[6]: For 4th param expected `Tensor` but got `int`.
        res = matern_kernel(a, b, nu=0.5, lengthscale=lengthscale)
        actual = (
            torch.tensor([[4, 2], [2, 0], [8, 6]], dtype=torch.float)
            .div_(-lengthscale)
            .exp()
        )
        self.assertLess(torch.norm(res - actual), 1e-3)
        # matern test 3/2
        # pyre-fixme[6]: For 4th param expected `Tensor` but got `int`.
        res = matern_kernel(a, b, nu=1.5, lengthscale=lengthscale)
        dist = torch.tensor([[4, 2], [2, 0], [8, 6]], dtype=torch.float).mul_(
            sqrt(3) / lengthscale
        )
        actual = (dist + 1).mul(torch.exp(-dist))
        self.assertLess(torch.norm(res - actual), 1e-3)
        # matern test 5/2
        # pyre-fixme[6]: For 4th param expected `Tensor` but got `int`.
        res = matern_kernel(a, b, nu=2.5, lengthscale=lengthscale)
        dist = torch.tensor([[4, 2], [2, 0], [8, 6]], dtype=torch.float).mul_(
            sqrt(5) / lengthscale
        )
        # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and `int`.
        actual = (dist**2 / 3 + dist + 1).mul(torch.exp(-dist))
        self.assertLess(torch.norm(res - actual), 1e-3)

        # test k(x,x) with no gradients
        # pyre-fixme[6]: For 4th param expected `Tensor` but got `float`.
        res = matern_kernel(b, b, nu=0.5, lengthscale=2.0)
        actual = (
            torch.tensor([[0, 2], [2, 0]], dtype=torch.float).div_(-lengthscale).exp()
        )
        self.assertLess(torch.norm(res - actual), 1e-3)

        # test unsupported nu
        with self.assertRaises(AxError):
            # pyre-fixme[6]: For 4th param expected `Tensor` but got `float`.
            matern_kernel(b, b, nu=0.0, lengthscale=2.0)

    def test_rbf_kernel(self) -> None:
        a = torch.tensor([4, 2, 8], dtype=torch.float).view(3, 1)
        b = torch.tensor([0, 2], dtype=torch.float).view(2, 1)
        lengthscale = 2
        # test rbf
        # pyre-fixme[6]: For 3rd param expected `Tensor` but got `int`.
        res = rbf_kernel(a, b, lengthscale=lengthscale)
        actual = (
            torch.tensor([[4, 2], [2, 0], [8, 6]], dtype=torch.float)
            .pow_(2.0)
            .mul_(-0.5 / (lengthscale**2))
            .exp()
        )
        self.assertLess(torch.norm(res - actual), 1e-3)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import warnings
from logging import Logger
from typing import Any, Iterable, List, Optional, Tuple

from ax.core.metric import Metric
from ax.utils.common.base import SortableBase
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import not_none

logger: Logger = get_logger(__name__)


class Objective(SortableBase):
    """Base class for representing an objective.

    Attributes:
        minimize: If True, minimize metric.
    """

    def __init__(self, metric: Metric, minimize: Optional[bool] = None) -> None:
        """Create a new objective.

        Args:
            metric: The metric to be optimized.
            minimize: If True, minimize metric. If None, will be set based on the
                `lower_is_better` property of the metric (if that is not specified,
                will raise a DeprecationWarning).

        """
        lower_is_better = metric.lower_is_better
        if minimize is None:
            if lower_is_better is None:
                warnings.warn(
                    f"Defaulting to `minimize=False` for metric {metric.name} not "
                    + "specifying `lower_is_better` property. This is a wild guess. "
                    + "Specify either `lower_is_better` on the metric, or specify "
                    + "`minimize` explicitly. This will become an error in the future.",
                    DeprecationWarning,
                )
                minimize = False
            else:
                minimize = lower_is_better
        if lower_is_better is not None:
            if lower_is_better and not minimize:
                warnings.warn(
                    f"Attempting to maximize metric {metric.name} with property "
                    "`lower_is_better=True`."
                )
            elif not lower_is_better and minimize:
                warnings.warn(
                    f"Attempting to minimize metric {metric.name} with property "
                    "`lower_is_better=False`."
                )
        self._metric = metric
        # pyre-fixme[4]: Attribute must be annotated.
        self.minimize = not_none(minimize)

    @property
    def metric(self) -> Metric:
        """Get the objective metric."""
        return self._metric

    @property
    def metrics(self) -> List[Metric]:
        """Get a list of objective metrics."""
        return [self._metric]

    @property
    def metric_names(self) -> List[str]:
        """Get a list of objective metric names."""
        return [m.name for m in self.metrics]

    def clone(self) -> Objective:
        """Create a copy of the objective."""
        return Objective(self.metric.clone(), self.minimize)

    def __repr__(self) -> str:
        return 'Objective(metric_name="{}", minimize={})'.format(
            self.metric.name, self.minimize
        )

    def get_unconstrainable_metrics(self) -> List[Metric]:
        """Return a list of metrics that are incompatible with OutcomeConstraints."""
        return self.metrics

    @property
    def _unique_id(self) -> str:
        return str(self)


class MultiObjective(Objective):
    """Class for an objective composed of a multiple component objectives.

    The Acquisition function determines how the objectives are weighted.

    Attributes:
        objectives: List of objectives.
    """

    weights: List[float]

    def __init__(
        self,
        objectives: Optional[List[Objective]] = None,
        **extra_kwargs: Any,  # Here to satisfy serialization.
    ) -> None:
        """Create a new objective.

        Args:
            objectives: The list of objectives to be jointly optimized.

        """
        # Support backwards compatibility for old API in which
        # MultiObjective constructor accepted `metrics` and `minimize`
        # rather than `objectives`
        if objectives is None:
            if "metrics" not in extra_kwargs:
                raise ValueError(
                    "Must either specify `objectives` or `metrics` "
                    "as input to `MultiObjective` constructor."
                )
            metrics = extra_kwargs["metrics"]
            minimize = extra_kwargs.get("minimize", False)
            warnings.warn(
                "Passing `metrics` and `minimize` as input to the `MultiObjective` "
                "constructor will soon be deprecated. Instead, pass a list of "
                "`objectives`. This will become an error in the future.",
                DeprecationWarning,
            )
            objectives = []
            for metric in metrics:
                lower_is_better = metric.lower_is_better or False
                _minimize = not lower_is_better if minimize else lower_is_better
                objectives.append(Objective(metric=metric, minimize=_minimize))

        # pyre-fixme[4]: Attribute must be annotated.
        self._objectives = not_none(objectives)

        # For now, assume all objectives are weighted equally.
        # This might be used in the future to change emphasis on the
        # relative focus of the exploration during the optimization.
        self.weights = [1.0 for _ in range(len(objectives))]

    @property
    def metric(self) -> Metric:
        """Override base method to error."""
        raise NotImplementedError(
            f"{type(self).__name__} is composed of multiple metrics"
        )

    @property
    def metrics(self) -> List[Metric]:
        """Get the objective metrics."""
        return [o.metric for o in self._objectives]

    @property
    def objectives(self) -> List[Objective]:
        """Get the objectives."""
        return self._objectives

    @property
    def objective_weights(self) -> Iterable[Tuple[Objective, float]]:
        """Get the objectives and weights."""
        return zip(self.objectives, self.weights)

    def clone(self) -> Objective:
        """Create a copy of the objective."""
        return MultiObjective(objectives=[o.clone() for o in self.objectives])

    def __repr__(self) -> str:
        return f"MultiObjective(objectives={self.objectives})"


class ScalarizedObjective(Objective):
    """Class for an objective composed of a linear scalarization of metrics.

    Attributes:
        metrics: List of metrics.
        weights: Weights for scalarization; default to 1.
    """

    weights: List[float]

    def __init__(
        self,
        metrics: List[Metric],
        weights: Optional[List[float]] = None,
        minimize: bool = False,
    ) -> None:
        """Create a new objective.

        Args:
            metric: The metric to be optimized.
            weights: The weights for the linear combination of metrics.
            minimize: If true, minimize the linear combination.

        """
        if weights is None:
            weights = [1.0 for i in range(len(metrics))]
        else:
            if len(weights) != len(metrics):
                raise ValueError("Length of weights must equal length of metrics")

        self._metrics = metrics
        self.weights = weights
        self.minimize = minimize

    @property
    def metric(self) -> Metric:
        """Override base method to error."""
        raise NotImplementedError(
            f"{type(self).__name__} is composed of multiple metrics"
        )

    @property
    def metrics(self) -> List[Metric]:
        """Get the metrics."""
        return self._metrics

    @property
    def metric_weights(self) -> Iterable[Tuple[Metric, float]]:
        """Get the metrics and weights."""
        return zip(self.metrics, self.weights)

    def clone(self) -> Objective:
        """Create a copy of the objective."""
        return ScalarizedObjective(
            metrics=[m.clone() for m in self.metrics],
            weights=self.weights.copy(),
            minimize=self.minimize,
        )

    def __repr__(self) -> str:
        return "ScalarizedObjective(metric_names={}, weights={}, minimize={})".format(
            [metric.name for metric in self.metrics], self.weights, self.minimize
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import abstractmethod

from ax.core.base_trial import TrialStatus

from ax.core.experiment import Experiment
from ax.utils.common.base import Base
from ax.utils.common.serialization import SerializationMixin


class CompletionCriterion(Base, SerializationMixin):
    """
    Simple class to descibe a condition which must be met for a GenerationStraytegy
    to move to its next GenerationStep.
    """

    def __init__(self) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def is_met(self, experiment: Experiment) -> bool:
        pass  # pragma: no cover


class MinimumTrialsInStatus(CompletionCriterion):
    def __init__(self, status: TrialStatus, threshold: int) -> None:
        self.status = status
        self.threshold = threshold

    def is_met(self, experiment: Experiment) -> bool:
        return len(experiment.trial_indices_by_status[self.status]) >= self.threshold


class MinimumPreferenceOccurances(CompletionCriterion):
    def __init__(self, metric_name: str, threshold: int) -> None:
        self.metric_name = metric_name
        self.threshold = threshold

    def is_met(self, experiment: Experiment) -> bool:
        data = experiment.fetch_data(metrics=[experiment.metrics[self.metric_name]])

        count_no = (data.df["mean"] == 0).sum()
        count_yes = (data.df["mean"] != 0).sum()

        return count_no >= self.threshold and count_yes >= self.threshold



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from logging import Logger
from typing import Dict, List

import numpy as np
from ax.core.observation import ObservationData
from ax.modelbridge.transforms.base import Transform
from ax.utils.common.logger import get_logger

logger: Logger = get_logger(__name__)


def ivw_metric_merge(
    obsd: ObservationData, conflicting_noiseless: str = "warn"
) -> ObservationData:
    """Merge multiple observations of a metric with inverse variance weighting.

    Correctly updates the covariance of the new merged estimates:
    ybar1 = Sum_i w_i * y_i
    ybar2 = Sum_j w_j * y_j
    cov[ybar1, ybar2] = Sum_i Sum_j w_i * w_j * cov[y_i, y_j]

    w_i will be infinity if any variance is 0. If one variance is 0., then
    the IVW estimate is the corresponding mean. If there are multiple
    measurements with 0 variance but means are all the same, then IVW estimate
    is that mean. If there are multiple measurements and means differ, behavior
    depends on argument conflicting_noiseless. "ignore" and "warn" will use
    the first of the measurements as the IVW estimate. "warn" will additionally
    log a warning. "raise" will raise an exception.

    Args:
        obsd: An ObservationData object
        conflicting_noiseless: "warn", "ignore", or "raise"
    """
    if len(obsd.metric_names) == len(set(obsd.metric_names)):
        return obsd
    if conflicting_noiseless not in {"warn", "ignore", "raise"}:
        raise ValueError(
            'conflicting_noiseless should be "warn", "ignore", or "raise".'
        )
    # Get indicies and weights for each metric.
    # weights is a map from metric name to a vector of the weights for each
    # measurement of that metric. indicies gives the corresponding index in
    # obsd.means for each measurement.
    weights: Dict[str, np.ndarray] = {}
    indicies: Dict[str, List[int]] = {}
    for metric_name in set(obsd.metric_names):
        indcs = [i for i, mn in enumerate(obsd.metric_names) if mn == metric_name]
        indicies[metric_name] = indcs
        # Extract variances for observations of this metric
        sigma2s = obsd.covariance[indcs, indcs]
        # Check for noiseless observations
        idx_noiseless = np.where(sigma2s == 0.0)[0]
        if len(idx_noiseless) == 0:
            # Weight is inverse of variance, normalized
            # Expected `np.ndarray` for 3rd anonymous parameter to call
            # `dict.__setitem__` but got `float`.
            # pyre-fixme[6]:
            weights[metric_name] = 1.0 / sigma2s
            weights[metric_name] /= np.sum(weights[metric_name])
        else:
            # Check if there are conflicting means for the noiseless observations
            means_noiseless = obsd.means[indcs][idx_noiseless]
            _check_conflicting_means(
                means_noiseless, metric_name, conflicting_noiseless
            )
            # The first observation gets all the weight.
            weights[metric_name] = np.zeros_like(sigma2s)
            weights[metric_name][idx_noiseless[0]] = 1.0
    # Compute the new values
    metric_names = sorted(set(obsd.metric_names))
    means = np.zeros(len(metric_names))
    covariance = np.zeros((len(metric_names), len(metric_names)))
    for i, metric_name in enumerate(metric_names):
        ys = obsd.means[indicies[metric_name]]
        means[i] = np.sum(weights[metric_name] * ys)
        # Calculate covariances with metric_name
        for j, metric_name2 in enumerate(metric_names[i:], start=i):
            for ii, idx_i in enumerate(indicies[metric_name]):
                for jj, idx_j in enumerate(indicies[metric_name2]):
                    covariance[i, j] += (
                        weights[metric_name][ii]
                        * weights[metric_name2][jj]
                        * obsd.covariance[idx_i, idx_j]
                    )
            covariance[j, i] = covariance[i, j]
    return ObservationData(
        metric_names=metric_names, means=means, covariance=covariance
    )


def _check_conflicting_means(
    means_noiseless: np.ndarray, metric_name: str, conflicting_noiseless: str
) -> None:
    if np.var(means_noiseless) > 0:
        message = f"Conflicting noiseless measurements for {metric_name}."
        if conflicting_noiseless == "warn":
            logger.warning(message)
        elif conflicting_noiseless == "raise":
            raise ValueError(message)


class IVW(Transform):
    """If an observation data contains multiple observations of a metric, they
    are combined using inverse variance weighting.
    """

    def _transform_observation_data(
        self,
        observation_data: List[ObservationData],
    ) -> List[ObservationData]:
        # pyre: conflicting_noiseless is declared to have type `str` but is
        # pyre-fixme[9]: used as type `typing.Union[float, int, str]`.
        conflicting_noiseless: str = self.config.get("conflicting_noiseless", "warn")
        return [
            ivw_metric_merge(obsd=obsd, conflicting_noiseless=conflicting_noiseless)
            for obsd in observation_data
        ]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional, TYPE_CHECKING

import numpy as np
from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ChoiceParameter, ParameterType
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.exceptions.core import UnsupportedError
from ax.modelbridge.transforms.base import Transform
from ax.models.types import TConfig

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import modelbridge as modelbridge_module  # noqa F401  # pragma: no cover


TRIAL_PARAM = "TRIAL_PARAM"


class TrialAsTask(Transform):
    """Convert trial to one or more task parameters.

    How trial is mapped to parameter is specified with a map like
    {parameter_name: {trial_index: level name}}.
    For example,
    {"trial_param1": {0: "level1", 1: "level1", 2: "level2"},}
    will create choice parameters "trial_param1" with is_task=True.
    Observations with trial 0 or 1 will have "trial_param1" set to "level1",
    and those with trial 2 will have "trial_param1" set to "level2". Multiple
    parameter names and mappings can be specified in this dict.

    The trial level mapping can be specified in config["trial_level_map"]. If
    not specified, defaults to a parameter with a level for every trial index.

    For the reverse transform, if there are multiple mappings in the transform
    the trial will not be set.

    Will raise if trial not specified for every point in the training data.

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: Optional[SearchSpace] = None,
        observations: Optional[List[Observation]] = None,
        modelbridge: Optional["modelbridge_module.base.ModelBridge"] = None,
        config: Optional[TConfig] = None,
    ) -> None:
        assert observations is not None, "TrialAskTask requires observations"
        # Identify values of trial.
        trials = {obs.features.trial_index for obs in observations}
        if isinstance(search_space, RobustSearchSpace):
            raise UnsupportedError(
                "TrialAsTask transform is not supported for RobustSearchSpace."
            )
        if None in trials:
            raise ValueError(
                "Unable to use trial as task since not all observations have "
                "trial specified."
            )
        # Get trial level map
        if config is not None and "trial_level_map" in config:
            # pyre: Attribute `trial_level_map` declared in class `ax.
            # pyre: modelbridge.transforms.trial_as_task.TrialAsTask` has type
            # pyre: `Dict[str, Dict[int, str]]` but is used as type `typing.
            # pyre-fixme[8]: Union[float, int, str]`.
            self.trial_level_map: Dict[str, Dict[int, str]] = config["trial_level_map"]
            # Validate
            for _p_name, level_dict in self.trial_level_map.items():
                # Check that trials match those in data
                level_map = set(level_dict.keys())
                if not trials.issubset(level_map):
                    raise ValueError(
                        f"Not all trials in data ({trials}) contained "
                        f"in trial level map for {_p_name} ({level_map})"
                    )
        else:
            # Set TRIAL_PARAM for each trial to the corresponding trial_index.
            # pyre-fixme[6]: Expected `Union[bytes, str, typing.SupportsInt]` for
            #  1st param but got `Optional[np.int64]`.
            self.trial_level_map = {TRIAL_PARAM: {int(b): str(b) for b in trials}}
        if len(self.trial_level_map) == 1:
            level_dict = next(iter(self.trial_level_map.values()))
            self.inverse_map: Optional[Dict[str, int]] = {
                v: k for k, v in level_dict.items()
            }
        else:
            self.inverse_map = None

    def transform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationFeatures]:
        for obsf in observation_features:
            if obsf.trial_index is not None:
                for p_name, level_dict in self.trial_level_map.items():
                    # pyre-fixme[6]: Expected `Union[bytes, str,
                    #  typing.SupportsInt]` for 1st param but got `Optional[np.int64]`.
                    obsf.parameters[p_name] = level_dict[int(obsf.trial_index)]
                obsf.trial_index = None
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        for p_name, level_dict in self.trial_level_map.items():
            level_values = sorted(set(level_dict.values()))
            if len(level_values) < 2:
                details = (
                    f"only 1 found: {level_values}" if level_values else "none found"
                )
                raise ValueError(
                    f"TrialAsTask transform expects 2+ task params, {details}"
                )
            trial_param = ChoiceParameter(
                name=p_name,
                parameter_type=ParameterType.STRING,
                # Expected `List[Optional[typing.Union[bool, float, str]]]` for 4th
                # parameter `values` to call
                # `ax.core.parameter.ChoiceParameter.__init__` but got
                # `List[str]`.
                # pyre-fixme[6]:
                values=level_values,
                is_ordered=False,
                is_task=True,
                sort_values=True,
            )
            search_space.add_parameter(trial_param)
        return search_space

    def untransform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationFeatures]:
        for obsf in observation_features:
            for p_name in self.trial_level_map:
                pval = obsf.parameters.pop(p_name)
            if self.inverse_map is not None:
                # pyre-fixme[61]: `pval` may not be initialized here.
                obsf.trial_index = np.int64(self.inverse_map[pval])
        return observation_features



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import logging

from logging import Logger
from typing import Dict, Iterable, List, NamedTuple, Union

import pandas as pd
from ax.core.map_data import MapKeyInfo
from ax.metrics.curve import AbstractCurveMetric
from ax.utils.common.logger import get_logger

logger: Logger = get_logger(__name__)

try:
    from tensorboard.backend.event_processing import (
        plugin_event_multiplexer as event_multiplexer,
    )
    from tensorboard.compat.proto import types_pb2

    logging.getLogger("tensorboard").setLevel(logging.CRITICAL)

    class TensorboardCurveMetric(AbstractCurveMetric):
        """A `CurveMetric` for getting Tensorboard curves."""

        # pyre-fixme[4]: Attribute must be annotated.
        MAP_KEY = MapKeyInfo(key="steps", default_value=0.0)

        @classmethod
        def get_curves_from_ids(
            cls, ids: Iterable[Union[int, str]]
        ) -> Dict[Union[int, str], Dict[str, pd.Series]]:
            """Get curve data from tensorboard logs.

            NOTE: If the ids are not simple paths/posix locations, subclass this metric
            and replace this method with an appropriate one that retrieves the log
            results.

            Args:
                ids: A list of string paths to tensorboard log directories.

            Returns:
                A dictionary mapping metric names to pandas Series of data.
            """
            return {idx: get_tb_from_posix(str(idx)) for idx in ids}

    def get_tb_from_posix(path: str) -> Dict[str, pd.Series]:
        r"""Get Tensorboard data from a posix path.

        Args:
            path: The posix path for the directory that contains the tensorboard logs.

        Returns:
            A dictionary mapping metric names to pandas Series of data.
        """
        logger.debug(f"Reading TB logs from {path}.")
        mul = event_multiplexer.EventMultiplexer(max_reload_threads=20)
        mul.AddRunsFromDirectory(path, None)
        mul.Reload()
        scalar_dict = mul.PluginRunToTagToContent("scalars")
        raw_result = [
            {"tag": tag, "event": mul.Tensors(run, tag)}
            for run, run_dict in scalar_dict.items()
            for tag in run_dict
        ]
        tb_run_data = {}
        for item in raw_result:
            latest_start_time = _get_latest_start_time(item["event"])
            steps = [e.step for e in item["event"] if e.wall_time >= latest_start_time]
            vals = [
                _get_event_value(e)
                for e in item["event"]
                if e.wall_time >= latest_start_time
            ]
            key = item["tag"]
            series = pd.Series(index=steps, data=vals).dropna()
            if any(series.index.duplicated()):  # pyre-ignore[16]
                # take average of repeated observations of the same "step"
                series = series.groupby(series.index).mean()  # pyre-ignore[16]
                logger.debug(
                    f"Found duplicate steps for tag {key}. "
                    "Removing duplicates by averaging."
                )
            tb_run_data[key] = series
        return tb_run_data

    # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
    #  `typing.List` to avoid runtime subscripting errors.
    def _get_latest_start_time(events: List) -> float:
        """In each directory, there may be previous training runs due to restarting
        training jobs.

        Args:
            events: A list of TensorEvents.

        Returns:
            The start time of the latest training run.
        """
        events.sort(key=lambda e: e.wall_time)
        start_time = events[0].wall_time
        for i in range(1, len(events)):
            # detect points in time where restarts occurred
            if events[i].step < events[i - 1].step:
                start_time = events[i].wall_time
        return start_time

    def _get_event_value(e: NamedTuple) -> float:
        r"""Helper function to check the dtype and then get the value
        stored in a TensorEvent."""
        tensor = e.tensor_proto  # pyre-ignore[16]
        if tensor.dtype == types_pb2.DT_FLOAT:
            return tensor.float_val[0]
        elif tensor.dtype == types_pb2.DT_DOUBLE:
            return tensor.double_val[0]
        elif tensor.dtype == types_pb2.DT_INT32:
            return tensor.int_val[0]
        else:
            raise ValueError(f"Tensorboard dtype {tensor.dtype} not supported.")

except ImportError:
    logger.warning(
        "tensorboard package not found. If you would like to use "
        "TensorboardCurveMetric, please install tensorboard."
    )
    pass
